# Copyright (C) Red Hat Inc.
#
# fedfind is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:   Adam Williamson <awilliam@redhat.com>

"""Library and CLI tool for locating and getting information about
Fedora composes.
"""

from __future__ import unicode_literals
from __future__ import print_function

__version__ = "4.0.0"

# vim: set textwidth=100 ts=8 et sw=4:
